# FUNCTION_my ensemble functions

#remapping cluster labels where needed
#еľǩǴ1ʼģmapɴ1ʼ
labels_remapping <- function(cluster_results){
  for (r in 1:nrow(cluster_results)){
    labels <- cluster_results[r,]
    if (max(labels) != length(levels(as.factor(labels)))){
      labels_new <- cbind(labels, t(t(c(1:length(labels)))))
      labels_inorder <- labels_new[order(labels_new[,1]),]
      last <- NULL
      curind <- 0
      for (i in 1:nrow(labels_inorder)){
        if (last != labels_inorder[i,1] || is.null(last)){
          last <- labels_inorder[i,1]
          curind <- curind + 1
        }
        labels_inorder[i,1] <- curind
      }
      labels_inorder <- labels_inorder[order(labels_inorder[,2]),]
      labels_new <- t(labels_inorder[,1])
      cluster_results[r,] <- labels_new
    }
  }
  return(cluster_results)
}

### Function hypergraph_generation performs transformation from clustering results to hypergraph
# object contains clustering results generated by all the single-cell clustering methods. nrow equals to No. of methods, and ncol is No. of cells
hypergraph_generation <- function(object){
  hypergraph <- matrix(data = 0, nrow = length(object), ncol = max(object))
  for (i in 1:length(hypergraph[1,])){
    hypergraph[,i] <- as.numeric(object == i)
  }
  return(hypergraph)
}


### Define function ints
ints <- function(s, tafter = 1000000000){
  tbefore <- sum(s)
  while (tafter/tbefore <= 0.5){
    tafter = tafter * 10
  }
  if (tbefore != 0){
    s = round(s * (tafter/tbefore))  # round()뵽ӽ
  }
  return(s)
}



### Function cspa performs cluster ensemble by cspa (CSPA)
# object is the clustering results generated by all single-cell clustering methods
# k is the maximum No. of clusters
cspa <- function(object, k = k){
  timestart<-Sys.time()
  hypergraph <- NULL
  print('Cluster ensembles using CSPA')
  
  for (i in 1:nrow(object)) {
    hypergraph_each <- hypergraph_generation(object[i,])
    hypergraph = cbind(hypergraph, hypergraph_each)
  }
  
  ### Function checks
  #֤ԾԽΪ1
  checks <- function(object){
    if (nrow(object) == ncol(object)){
      if (sum(diag(object) != 1) > 0){
        print ('self-similarity object[i,i] is not always 1')
        for (i in 1:nrow(object)){
          object[i,i] <- 1;
        }
        print ('diagonal is set to 1')
      }
    }
    return(object)
  }
  
  similarity_matrix <- checks((hypergraph %*% t(hypergraph))/nrow(object))
  
  SNF_SpectralClustering_clusters <- c(SNFtool::spectralClustering(similarity_matrix, k, type = 3))
  #row.names(SNF_SpectralClustering_clusters) <- "SNF_SpectralClustering_clusters"
  source("FUNCTION_my SNF spectral clustering.R")
  my_SpectralClustering_clusters <- c(spectral_clustering(similarity_matrix, k, normal = T))
  #row.names(my_SpectralClustering_clusters) <- "my_SpectralClustering_clusters"
  
  
  timeend<-Sys.time()
  runningtime<-timeend-timestart
  print(timestart)
  print(timeend)
  cspa_time = paste('Runing time of CSPA is ', runningtime, sep = "")
  print(cspa_time)
  
  cluster_ensembles <- NULL
  
  cluster_ensembles <- rbind(cluster_ensembles, matrix(SNF_SpectralClustering_clusters, nrow = 1, byrow = TRUE))
  cluster_ensembles <- rbind(cluster_ensembles, matrix(my_SpectralClustering_clusters, nrow = 1, byrow = TRUE))
  #cluster_ensembles$CSPA_SNF_SpectralClustering <- c(SNF_SpectralClustering_clusters)
  #cluster_ensembles$CSPA_my_SpectralClustering <- c(my_SpectralClustering_clusters)
  return(cluster_ensembles)
}


### Function check can validate and fix problems in cluster labels
# labels are the ensemble cluster lables generated by function sgraph.
# 
check <- function(labels){
  labels_new <- cbind(t(labels), t(t(c(1:length(labels)))))
  labels_inorder <- labels_new[order(labels_new[,1]),]
  
  last <- NULL
  curind <- 0
  for (i in 1:nrow(labels_inorder)){
    if (last != labels_inorder[i,1] || is.null(last)){
      last <- labels_inorder[i,1]
      curind <- curind + 1
    }
    labels_inorder[i,1] <- curind
  }
  labels_inorder <- labels_inorder[order(labels_inorder[,2]),]
  labels_new <- t(labels_inorder[,1])
  
  if (max(labels) != max(labels_new)){
    print ('Result checking: not a dense integer mapping')
    labels <- NULL
    labels <- labels_new
    print (paste('Remapped to 1 to ', max(labels), ' clusters', sep = ''))
  }
  return(labels)
}
















